# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2010/12/21
# copy: (C) Copyright 2010-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import sys
import argparse
import BaseHTTPServer
import threading
from aadict import aadict
import morph
import blessings
import atexit
import logging

from .engine import parseSedExpression, ReplayServer, \
    MultiLogger, StreamLogger, DisplayLogger, LoggingRequestHandler
from .i18n import _

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

class LogFmt(logging.Formatter):
  def lvlstr(self, level):
    if level >= logging.CRITICAL : return '[**] CRITICAL:'
    if level >= logging.ERROR    : return '[**] ERROR:'
    if level >= logging.WARNING  : return '[++] WARNING:'
    if level >= logging.INFO     : return '[--] INFO:'
    # if level >= logging.DEBUG    : return '[  ] DEBUG:'
    return '[  ]'
  def format(self, record):
    msg = record.getMessage()
    #pfx = '%s|%s: ' % (self.levelString[record.levelno], record.name)
    pfx = self.lvlstr(record.levelno) + ' '
    return pfx + ('\n' + pfx).join(msg.split('\n'))

def LogRecord_getMessage_i18n(self):
  tmp = self.args
  self.args = None
  msg = self._real_getMessage()
  self.args = tmp
  if tmp:
    return _(msg) % tmp
  return _(msg)

logging.LogRecord._real_getMessage = logging.LogRecord.getMessage
logging.LogRecord.getMessage = LogRecord_getMessage_i18n

#------------------------------------------------------------------------------
def breaklines(options):
  while True:
    sys.stdin.read()
    print '\r\n' * options.breaklines

#------------------------------------------------------------------------------
def getDefaultMarkup(color):
  term = blessings.Terminal(force_styling=True)
  markup = aadict(
    packet           = term.bold_magenta,
    rline            = term.bold,
    headerName       = lambda msg: term.normal() + msg,
    headerValue      = lambda msg: term.normal() + msg,
    requestPrefix    = term.red,
    responsePrefix   = term.green,
    )
  if not color:
    for key in markup.keys():
      markup[key] = lambda msg: msg
  return markup

#------------------------------------------------------------------------------
def main(argv=None):

  cli = argparse.ArgumentParser(
    description='A simple, logging, colorizing, prettifying, HTTP proxy.',
  )

  cli.add_argument(
    _('-v'), _('--verbose'),
    dest='verbose', default=0, action='count',
    help=_('enable verbose output to STDERR (multiple'
           ' invocations increase verbosity)'))

  # cli.add_argument(
  #   _('--ssl'),
  #   dest='ssl', default=False, action='store_true',
  #   help=_('upgrade outbound requests to HTTPS'))

  cli.add_argument(
    _('-p'), _('--port'), metavar=_('PORT'),
    dest='local', default=80, action='store', type=int,
    help=_('specify the local port to listen on for new'
           ' connections'))

  cli.add_argument(
    _('-r'), _('--remote'), metavar=_('HOSTPORT'),
    dest='remote', default=None, action='store',
    help=_('specify the remote server and optionally port (e.g.'
           ' "host" or "host:port")'))

  # cli.add_argument(
  #   _('-H'), _('--rewrite-host'),
  #   dest='rewriteHost', default=False, action='store_true',
  #   help=_('(TODO: IMPLEMENT) rewrite the outbound HTTP "Host"'
  #          ' header, if present, to that specified by --remote'))

  cli.add_argument(
    _('-s'), _('--response-sed'), metavar=_('SED-EXPR'),
    dest='responseSed', default=[], action='append',
    help=_('use sed-like transformation of the response body'))

  cli.add_argument(
    _('-o'), _('--output'), metavar=_('FILENAME'),
    dest='outfile', default=None, action='store',
    help=_('store all transactions to the specified file'
           ' in binary format so that it can later be replayed'
           ' with --input'))

  cli.add_argument(
    _('-a'), _('--append'), metavar=_('FILENAME'),
    dest='appfile', default=None, action='store',
    help=_('same as "--output", but append to the current'
           ' content, if the file exists'))

  cli.add_argument(
    _('-i'), _('--input'), metavar=_('FILENAME'),
    dest='infile', default=None, action='store',
    help=_('replay previous transactions recorded with'
           ' "--output" (implies "--display")'))

  cli.add_argument(
    _('-d'), _('--display'),
    dest='display', default=False, action='store_true',
    help=_('display all transactions to STDOUT in human-'
           'readable form'))

  cli.add_argument(
    _('-c'), _('--color'),
    dest='color', default=False, action='store_true',
    help=_('use terminal color codes to differentiate HTTP'
           ' headers, content, requests and responses'
           ' (implies "--display")'))

  cli.add_argument(
    _('-t'), _('--theme'), metavar=_('NAME'),
    dest='theme', default=None, action='store',
    help=_('overrides the default color theme; can be any of the'
           ' styles supported by Pygments (implies "--color")'))

  cli.add_argument(
    _('-u'), _('--uncompress'),
    dest='uncompress', default=False, action='store_true',
    help=_('uncompress body data (if compressed) before'
           ' displaying it (implies "--display")'))

  cli.add_argument(
    _('-f'), _('--format'),
    dest='format', default=False, action='store_true',
    help=_('turns on format prettification of known formats'
           ' including XML, YAML, and JSON (implies'
           ' "--uncompress" and "--display")'))

  cli.add_argument(
    _('-b'), _('--breaklines'), metavar=_('COUNT'),
    dest='breaklines', default=60, type=int,
    help=_('number of lines to output to STDOUT when ^D'
           ' (Control-D) is pressed'))

  #----------------------------------------------------------------------------

  options = cli.parse_args(argv)

  if not options.remote and not options.infile:
    cli.error(_('please specify the remote server ("--remote") or input'
                ' file ("--input")'))

  if options.appfile and options.outfile:
    cli.error(_('only one of "--output" or "--append" can be specified'))

  if options.remote and options.infile:
    cli.error(_('only one of "--remote" or "--input" can be specified'))

  options.responseSed = [parseSedExpression(expr)
                         for expr in options.responseSed]

  # convert options to a dict
  options = aadict(morph.omit(options))

  #----------------------------------------------------------------------------

  rootlog = logging.getLogger()
  rootlog.setLevel(logging.ERROR)
  handler = logging.StreamHandler()
  handler.setFormatter(LogFmt())
  rootlog.addHandler(handler)
  if options.verbose == 1    : rootlog.setLevel(logging.INFO)
  elif options.verbose == 2  : rootlog.setLevel(logging.DEBUG)
  elif options.verbose > 2   : rootlog.setLevel(1)

  #----------------------------------------------------------------------------

  if options.infile:
    options.display = True
    if options.infile == '-':
      options.infile = sys.stdin
      log.info('replaying transactions from <STDIN>')
    else:
      log.info('replaying transactions from "%s"', options.infile)
      options.infile = open(options.infile, 'rb')
      atexit.register(options.infile.close)
    server = ReplayServer(options, options.infile)
  else:
    # port = options.ssl and 443 or 80
    port = 80
    if options.remote.find(':') >= 0:
      options.remote, port = options.remote.split(':', 1)
      port = int(port)
    options.remote = (options.remote, port)
    options.local = ('localhost', options.local)
    server = BaseHTTPServer.HTTPServer(options.local, LoggingRequestHandler)
    log.info('accepting connections on %s:%r', *options.local)

  server.options = options
  server.logger  = MultiLogger()

  if options.format:
    options.uncompress = True

  if options.theme:
    options.color = True

  if options.color or options.uncompress:
    options.display = True

  options.markup = getDefaultMarkup(options.color)

  append = False
  if options.appfile:
    options.outfile = options.appfile
    append = True

  if options.outfile:
    if options.outfile == '-':
      server.logger.loggers.append(StreamLogger(sys.stdout, options=options))
    else:
      fstream = open(options.outfile, 'ab' if append else 'wb')
      server.logger.loggers.append(StreamLogger(fstream, options=options))
      atexit.register(fstream.close)

  if options.display:
    server.logger.loggers.append(DisplayLogger(sys.stdout, options=options))

  if not ( options.infile or options.outfile or options.display ):
    log.warn('no logging/displaying configured - acting as a simple proxy')

  try:
    if not options.infile:
      t = threading.Thread(target=breaklines, args=(options,))
      t.daemon = True
      t.start()
    server.serve_forever()
  except KeyboardInterrupt:
    log.info('aborted by user - exiting')

  return 0

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@metagriffin.net>
# date: 2014/02/27
# copy: (C) Copyright 2014-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import six
import yaml
import morph
import json

#------------------------------------------------------------------------------
def _writeObject(obj, out, color=True, level=0, indent='  '):
  def _write(obj, level):
    curdent = indent * level
    if morph.isseq(obj):
      out.write('[\n')
      last = len(obj)
      for idx, val in enumerate(obj):
        out.write(curdent)
        out.write(indent)
        _write(val, level=level + 1)
        if idx + 1 < last:
          out.write(',')
        out.write('\n')
      out.write(curdent)
      out.write(']')
      return
    if morph.isdict(obj):
      out.write('{\n')
      last = len(obj)
      for idx, (key, val) in enumerate(obj.items()):
        out.write(curdent)
        out.write(indent)
        _write(key, level=level + 1)
        out.write(': ')
        _write(val, level=level + 1)
        if idx + 1 < last:
          out.write(',')
        out.write('\n')
      out.write(curdent)
      out.write('}')
      return
    out.write(json.dumps(obj))
  return _write(obj, level)

#------------------------------------------------------------------------------
def prettify(input, output=None, strict=True, color=True):
  stream = output or six.StringIO()
  if morph.isstr(input) or isinstance(input, six.binary_type):
    data = input
  else:
    data = input.read()
  try:
    ydata = yaml.load(data)
  except Exception as err:
    if strict:
      raise
    if output:
      output.write(data)
      return
    return data
  _writeObject(ydata, stream, color=True)
  if output:
    return
  return stream.getvalue()

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/12/14
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import unittest
import six
from aadict import aadict

from .engine import parseSedExpression, ReplayServer, \
    MultiLogger, StreamLogger, DisplayLogger, LoggingRequestHandler
from .cli import getDefaultMarkup

#------------------------------------------------------------------------------
class TestProxylog(unittest.TestCase):

  maxDiff = None

  #----------------------------------------------------------------------------
  def test_input_xml(self):
    yaml = '''\
---
content: null
headers:
  host: host.example.com:80
meta:
  client: 127.0.0.1:1234
  isRequest: true
  processID: 1
  requestID: 1
  server: 127.0.0.1:80
  ts: 1234567890.0
rline: GET /path/to/data.xml HTTP/1.1
---
content: '<root ><node attr="value"/></root>'
headers:
  content-length: '34'
  content-type: application/xml
  date: Fri, 13 Feb 2009 23:31:30 GMT
  last-modified: Fri, 13 Feb 2009 23:31:30 GMT
  server: SimpleHTTP/0.6 Python/2.7.3
meta:
  client: 127.0.0.1:1234
  isRequest: false
  processID: 1
  requestID: 1
  server: 127.0.0.1:80
  ts: 1234567890.1
rline: HTTP/1.? 200 OK
'''
    chk = u'''\
[1m[35m[1234567890.000] 127.0.0.1:1234 --> 127.0.0.1:80 (00000001.00000001)(B[m
[31m  00000001.00000001 > (B[m[1mGET /path/to/data.xml HTTP/1.1(B[m
[31m  00000001.00000001 > (B[m(B[mHost: (B[mhost.example.com:80
[31m  00000001.00000001 > (B[m
[1m[35m[1234567890.100] 127.0.0.1:1234 <-- 127.0.0.1:80 (00000001.00000001)(B[m
[32m  00000001.00000001 < (B[m[1mHTTP/1.? 200 OK(B[m
[32m  00000001.00000001 < (B[m(B[mDate: (B[mFri, 13 Feb 2009 23:31:30 GMT
[32m  00000001.00000001 < (B[m(B[mLast-Modified: (B[mFri, 13 Feb 2009 23:31:30 GMT
[32m  00000001.00000001 < (B[m(B[mContent-Length: (B[m34
[32m  00000001.00000001 < (B[m(B[mContent-Type: (B[mapplication/xml
[32m  00000001.00000001 < (B[m(B[mServer: (B[mSimpleHTTP/0.6 Python/2.7.3
[32m  00000001.00000001 < (B[m
[32m  00000001.00000001 < (B[m[38;5;30m<?xml version="1.0" encoding="UTF-8"?>[39m
[32m  00000001.00000001 < (B[m[38;5;90;01m<root[39;00m[38;5;90;01m>[39;00m
[32m  00000001.00000001 < (B[m  [38;5;90;01m<node[39;00m [38;5;64mattr=[39m[38;5;167m"value"[39m[38;5;90;01m/>[39;00m
[32m  00000001.00000001 < (B[m[38;5;90;01m</root>[39;00m
[32m  00000001.00000001 < (B[m
'''
    options = aadict(
      color     = True,
      format    = True,
      infile    = six.StringIO(yaml),
      output    = six.StringIO(),
      markup    = getDefaultMarkup(True),
    )
    server = ReplayServer(options, options.infile)
    server.options = options
    server.logger  = DisplayLogger(options.output, options=options)
    server.serve_forever()
    self.assertMultiLineEqual(options.output.getvalue(), chk)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2010/12/21
# copy: (C) Copyright 2010-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import os
import sys
import six
import gzip
import BaseHTTPServer
# todo: replace with six.urllib...
import urllib2
import time
import yaml
import re
import pxml
from aadict import aadict
from requests.structures import CaseInsensitiveDict as idict
import pygments
import pygments.lexers
import pygments.formatters
import json
import morph
import logging

from .i18n import _
from . import pyaml

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
def mkdirp(path):
  if os.path.isdir(path):
    return
  os.makedirs(path)

#------------------------------------------------------------------------------
def gunzip(data):
  if data is None:
    return data
  return gzip.GzipFile(fileobj=six.StringIO(data)).read()

#------------------------------------------------------------------------------
def headerCase(h):
  return '-'.join([e.capitalize() for e in h.split('-')])

#------------------------------------------------------------------------------
class Logger(object):
  def __init__(self, options=None):
    self.options = aadict(
      showIndent       = True,
      showPacket       = True,
      showRequestLine  = True,
      showHeaders      = True,
      showRaw          = False,
      ).update(options or [])
  def logMessage(self, messageInfo, rLine, headers, body):
    raise NotImplementedError()
  def close(self):
    pass

#------------------------------------------------------------------------------
def getLexer(content, contentType):
  try:
    if contentType and ( '/yaml' in contentType or '+yaml' in contentType ):
      contentType = 'text/x-yaml'
    return pygments.lexers.get_lexer_for_mimetype(contentType)
  except Exception:
    try:
      return pygments.lexers.guess_lexer(content)
    except Exception:
      return None

#------------------------------------------------------------------------------
def formatify_xml(content, contentType):
  # todo: use the color scheme from `theme`...
  output = six.StringIO()
  pxml.prettify(six.StringIO(content.strip()), output, strict=False, color=False)
  return output.getvalue() or content

#------------------------------------------------------------------------------
def u2s(obj):
  if six.PY2 and isinstance(obj, unicode):
    return obj.encode('utf-8')
  if morph.isseq(obj):
    return [u2s(el) for el in obj]
  if morph.isdict(obj):
    return {u2s(k): u2s(v) for k, v in obj.items()}
  return obj

#------------------------------------------------------------------------------
def formatify_json(content, contentType):
  try:
    return formatify_yaml(yaml.dump(u2s(json.loads(content))), 'application/yaml')
  except Exception as err:
    return content

#------------------------------------------------------------------------------
def formatify_yaml(content, contentType):
  return pyaml.prettify(content, strict=False, color=False)

#------------------------------------------------------------------------------
def formatify(content, contentType):
  lexer = getLexer(content, contentType)
  log.debug('formatifying content-type %r (lexer %r)', contentType, lexer)
  if isinstance(lexer, (pygments.lexers.XmlLexer, pygments.lexers.LassoXmlLexer)) \
      or ( contentType and ( '/xml' in contentType or '+xml' in contentType ) ):
    return formatify_xml(content, contentType)
  if isinstance(lexer, pygments.lexers.JsonLexer):
    return formatify_json(content, contentType)
  if isinstance(lexer, pygments.lexers.YamlLexer):
    return formatify_yaml(content, contentType)
  return content

#------------------------------------------------------------------------------
def colorize(content, contentType, theme=None):
  lexer = getLexer(content, contentType)
  formatter = pygments.formatters.Terminal256Formatter(style=theme or 'perldoc')
  result = pygments.highlight(content, lexer, formatter)
  return result or content

#------------------------------------------------------------------------------
class DisplayLogger(Logger):
  def __init__(self, stream, *args, **kw):
    super(DisplayLogger, self).__init__(*args, **kw)
    self.stream = stream
  def logMessage(self, msg, rLine, headers, body):
    if self.options.showIndent:
      indent = self.options.markup['requestPrefix' if msg.isRequest else 'responsePrefix'](
        '  {:08x}.{:08x} {} '.format(
          msg.processID, msg.requestID, '>' if msg.isRequest else '<'))
    else:
      indent = ''
    content = body
    if content and self.options.uncompress \
       and headers.get('content-encoding') == 'gzip':
      content = gunzip(content)
    if content and self.options.format:
      content = formatify(content, headers.get('content-type'))
    if content and self.options.color:
      content = colorize(
        content, headers.get('content-type'), theme=self.options.theme)
    if self.options.showPacket:
      print >>self.stream, self.options.markup.packet(
        '[{:0.3f}] {}:{} {} {}:{} ({:08x}.{:08x})'.format(
          msg.ts,
          msg.client[0], msg.client[1],
          '-->' if msg.isRequest else '<--',
          msg.server[0], msg.server[1],
          msg.processID, msg.requestID,
          ))
    if self.options.showRequestLine:
      print >>self.stream, indent + self.options.markup.rline(rLine)
    if self.options.showHeaders:
      for k,v in headers.items():
        print >>self.stream, indent + '{}: {}'.format(
          self.options.markup.headerName(headerCase(k)),
          self.options.markup.headerValue(v))
      print >>self.stream, indent
    if content is not None:
      if self.options.showRaw:
        self.stream.write(content)
      else:
        print >>self.stream, indent + ('\n' + indent).join(content.split('\n'))
    self.stream.flush()

#------------------------------------------------------------------------------
class MultiLogger(Logger):
  def __init__(self, loggers=None, *args, **kw):
    super(MultiLogger, self).__init__(*args, **kw)
    self.loggers = loggers or []
  def logMessage(self, messageInfo, rLine, headers, body):
    for logger in self.loggers:
      logger.logMessage(messageInfo, rLine, headers, body)
  def close(self):
    for logger in self.loggers:
      logger.close()

#------------------------------------------------------------------------------
class LoggingRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

  REQUESTNUM = 0

  #----------------------------------------------------------------------------
  def __init__(self, request, client_address, server):
    self.options = server.options
    self.logger  = server.logger
    LoggingRequestHandler.REQUESTNUM += 1
    self.rid = LoggingRequestHandler.REQUESTNUM
    self.pid = os.getpid()
    log.info('connection from client: %s:%r', *client_address)
    BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, request, client_address, server)

  #----------------------------------------------------------------------------
  def do_GET(self):
    self.logRequest(None)
    self.sendRequest(None)
    self.logger.close()

  #----------------------------------------------------------------------------
  def do_POST(self):
    data = self.rfile.read(int(self.headers.get('content-length','0')))
    self.logRequest(data)
    self.sendRequest(data, method='POST')
    self.logger.close()

  #----------------------------------------------------------------------------
  def do_PUT(self):
    data = self.rfile.read(int(self.headers.get('content-length','0')))
    self.logRequest(data)
    self.sendRequest(data, method='PUT')
    self.logger.close()

  #----------------------------------------------------------------------------
  def sendRequest(self, content, method='GET'):
    req = urllib2.Request('%s://%s:%d%s' % (self.options.ssl and 'https' or 'http',
                                            self.options.remote[0],
                                            self.options.remote[1],
                                            self.path),
                          None, self.headers)

    # not using the build_opener as i want to handle 302s...
    #   opener  = urllib2.build_opener()
    opener = urllib2.OpenerDirector()
    opener.add_handler(urllib2.HTTPSHandler())
    opener.add_handler(urllib2.HTTPHandler())

    req.add_header('Host', '%s:%d' % (self.options.remote[0], self.options.remote[1]))

    if method not in 'GET':
      req.get_method = lambda: method

    if content is not None:
      resp = opener.open(req, content)
    else:
      resp = opener.open(req)

    info = resp.info()
    data = resp.read()

    def sedform(data):
      for xform in self.options.responseSed:
        data = xform(data)
      return data

    data = sedform(data)
    headers = idict({k: sedform(v) for k, v in info.items()})

    self.logResponse(resp, headers, data)
    self.sendResponse(resp, headers, data)

  #----------------------------------------------------------------------------
  def sendResponse(self, resp, info, data):
    # todo: is there really no way of determining the response HTTP version???
    self.wfile.write('%s %s %s\r\n' % ('HTTP/1.0', resp.code, resp.msg))
    self.wfile.write('Connection: close\r\n')
    for k,v in info.items():
      if k in ('connection', 'transfer-encoding'): continue
      # todo: encode headers in any way?
      self.wfile.write('%s: %s\r\n' % (headerCase(k), v))
    self.wfile.write('\r\n')
    self.wfile.write(data)

  #----------------------------------------------------------------------------
  def logRequest(self, content):
    self.logger.logMessage(aadict(
      ts        = time.time(),
      requestID = self.rid,
      processID = self.pid,
      isRequest = True,
      client    = (self.client_address[0], self.client_address[1]),
      server    = (self.options.remote[0], self.options.remote[1]),
      ), '%s %s %s' % (self.command, self.path, self.request_version),
      self.headers, content)

  #----------------------------------------------------------------------------
  def logResponse(self, response, headers, content):
    self.logger.logMessage(aadict(
      ts        = time.time(),
      requestID = self.rid,
      processID = self.pid,
      isRequest = False,
      client    = (self.client_address[0], self.client_address[1]),
      server    = (self.options.remote[0], self.options.remote[1]),
      ), '%s %s %s' % ('HTTP/1.?', response.code, response.msg),
      # todo: is there really no way of determining the response HTTP version???
      headers, content)

#------------------------------------------------------------------------------
class StreamLogger(Logger):
  def __init__(self, stream, *args, **kw):
    super(StreamLogger, self).__init__(*args, **kw)
    self.stream = stream
  def logMessage(self, msg, rLine, headers, body):
    self.stream.write('---\n')
    meta = dict(msg)
    if 'client' in meta:
      meta['client'] = ':'.join([str(el) for el in meta['client']])
    if 'server' in meta:
      meta['server'] = ':'.join([str(el) for el in meta['server']])
    data = dict(meta=meta, rline=str(rLine), headers=dict(headers), content=body)
    yaml.dump(data, self.stream, default_flow_style=False)
    self.stream.write('\n')

#------------------------------------------------------------------------------
class ReplayServer(object):
  def __init__(self, options, stream):
    self.options  = options
    self.stream   = stream
  def serve_forever(self):
    for record in yaml.load_all(self.stream):
      msg  = aadict(record['meta'])
      hdrs = idict(record['headers'])
      if 'client' in msg:
        msg['client'] = msg['client'].split(':')
        msg['client'][1] = int(msg['client'][1])
      if 'server' in msg:
        msg['server'] = msg['server'].split(':')
        msg['server'][1] = int(msg['server'][1])
      self.logger.logMessage(msg, record['rline'], hdrs, record['content'])

#------------------------------------------------------------------------------
# TODO: replace with `reparse` once it has been converted to a library...
def parseSedExpression(expr):
  if expr[0] != 's':
    raise NotImplementedError(_('only sed operator "s" is functional'))
  expr = expr.split(expr[1])
  if len(expr) != 4:
    raise TypeError(_('invalid sed "s" expression'))
  flags = 0
  for c in expr[3].upper():
    flags |= getattr(re, c, 0)
  regex = re.compile(expr[1], flags=flags)
  def xform(data):
    return regex.sub(expr[2], data)
  return xform

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
